import sys
from argparse import ArgumentParser

import yaml
from miniscule.base import read_config, load_config
from miniscule.logs import init_logging


def _create_parser():
    parser = ArgumentParser()
    parser.add_argument("path", metavar="PATH", nargs="?")
    parser.add_argument("-c", "--command", metavar="STR")
    return parser


def main():
    init_logging()
    parser = _create_parser()
    (settings, _) = parser.parse_known_args()
    if settings.command:
        obj = load_config(settings.command)
    else:
        obj = read_config(settings.path)

    if isinstance(obj, str):
        print(obj)
    else:
        yaml.dump(obj, sys.stdout)



import json
import logging
import os

import smart_open as smart
import yaml

from miniscule.exceptions import MinisculeError

log = logging.getLogger(__name__)


class ConfigLoader(yaml.SafeLoader):
    """Loader that resolves all miniscule tags.  Derives from
    :class:`yaml.SafeLoader` in `PyYAML <https://pypi.org/project/PyYAML/>`_.
    """

    # pylint: disable=too-many-ancestors
    def __init__(self, stream):
        super(ConfigLoader, self).__init__(stream)
        yaml.add_constructor("!or", or_constructor, Loader=ConfigLoader)
        yaml.add_constructor("!env", env_constructor, Loader=ConfigLoader)
        yaml.add_constructor("!merge", merge_constructor, Loader=ConfigLoader)
        yaml.add_constructor(
            "!include/json", include_json_constructor, Loader=ConfigLoader
        )
        yaml.add_constructor("!include", include_constructor, Loader=ConfigLoader)


def or_constructor(loader, node):
    for expr in loader.construct_sequence(node):
        if expr is not None:
            return expr
    return None


def env_constructor(loader, node):
    name = loader.construct_yaml_str(node)
    if name in os.environ:
        return yaml.load(os.getenv(name), Loader=loader.__class__)
    return None


def merge_constructor(loader, node):
    result = {}
    for m in loader.construct_sequence(node, deep=True):
        if isinstance(m, dict):
            result.update(m)
        elif m is None:
            pass
        else:
            raise MinisculeError("merge", "Arguments should be maps")
    return result


def _yaml_reader(loader, path, stream):
    # pylint: disable=unused-argument
    try:
        return yaml.load(stream, Loader=loader.__class__)
    except yaml.parser.ParserError as exc:
        raise MinisculeError("include", f"Invalid yaml in {path}") from exc


def _json_reader(loader, path, stream):
    # pylint: disable=unused-argument
    try:
        return json.load(stream)
    except json.JSONDecodeError as exc:
        raise MinisculeError("include", f"Invalid json in {path}") from exc


def _default_reader(loader, path, stream):
    # pylint: disable=unused-argument
    return stream.read().strip()


def include_constructor(loader, node):
    readers = {".yaml": _yaml_reader, ".yml": _yaml_reader, ".json": _json_reader}

    path = loader.construct_yaml_str(node)
    _, ext = os.path.splitext(path)

    reader = readers.get(ext, _default_reader)
    try:
        with smart.open(path, "r") as stream:
            return reader(loader, path, stream)
    except Exception as exc:
        log.warning("Failed to open file at %s", path)
        raise MinisculeError("include", str(exc))


def include_json_constructor(loader, node):
    path = loader.construct_yaml_str(node)
    try:
        with open(path, "r") as h:
            return json.load(h)
    except FileNotFoundError:
        log.warning("No file at %s", path)
    except json.JSONDecodeError:
        log.warning("Invalid JSON in file %s", path)
    return None


def load_config(stream, Loader=ConfigLoader):
    """Read configuration from a stream.

    :param stream: The stream to read from.
    :param Loader: The loader to use.  This allows clients to extend the list of
        tags that can be resolved.

    :returns: The parsed YAML in which the tags known to :code:`Loader` have been
              resolved.
    """
    return yaml.load(stream, Loader)


def read_config(path=None, Loader=ConfigLoader):
    """Read configuration from a file.

    :param path: Path of the file from which to read the configuration.  If
        None, the path is read from the value of the ``CONFIG`` environment
        variable. If no such variable, path defaults to ``config.yaml``.
    :param Loader: See :func:`miniscule.load_config`.

    :returns: See :func:`miniscule.load_config`.
    """
    path = path or os.environ.get("CONFIG", "config.yaml")
    with open(path, "r") as stream:
        return load_config(stream, Loader=Loader)



from yaml.parser import ParserError


class MinisculeError(ParserError):
    def __init__(self, tag, message):
        super().__init__()
        self.tag = tag
        self.message = message

    def __repr__(self):
        return "<{}(!{}, {})>".format(self.__class__.__name__, self.tag, self.message)



# pylint: disable=wildcard-import, unused-import
from miniscule.base import *
from miniscule.logs import *
from miniscule.main import main
import miniscule.aws

__all__ = ["read_config", "load_config", "init_logging", "main", "init"]



import logging
import logging.config
import asyncio

try:
    # Python 3.7 and newer, fast reentrant implementation
    # witohut task tracking (not needed for that when logging)
    from queue import SimpleQueue as Queue
except ImportError:
    from queue import Queue
from typing import List

import yaml

from miniscule.base import read_config, ConfigLoader

log = logging.getLogger(__name__)


class LocalQueueHandler(logging.handlers.QueueHandler):
    def emit(self, record: logging.LogRecord) -> None:
        # Removed the call to self.prepare(), handle task cancellation
        try:
            self.enqueue(record)
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)


def setup_logging_queue() -> None:
    """Move log handlers to a separate thread.

    Replace handlers on the root logger with a LocalQueueHandler, and start a
    logging.QueueListener holding the original handlers.
    """

    queue = Queue()  # type: ignore
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handlers: List[logging.Handler] = []

    handler = LocalQueueHandler(queue)
    root.addHandler(handler)
    for h in root.handlers[:]:
        if h is not handler:
            root.removeHandler(h)
            handlers.append(h)

    listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    listener.start()


def init_logging(config=None, key="log_config", asynchronous=False):
    """Initialize logging.

    :param config: The configuration dictionary
    :param key: The key under which the path of the logging configuration is
        stored.

    :returns: Nothing
    """
    path = None
    try:
        path = (config or read_config()).get(key)
    except FileNotFoundError:
        pass
    if path is None:
        log.debug("No logging configuration specified")
        return None

    with open(path, "r") as handle:
        log_config = yaml.load(handle.read(), Loader=yaml.SafeLoader)
        logging.config.dictConfig(log_config)

    if asynchronous:
        setup_logging_queue()

    return log_config


def _read_logging_config(path: str):
    with open(path, "r") as stream:
        return yaml.safe_load(stream.read())


def _set_up_logging(logging_config, asynchronous=False):
    logging.config.dictConfig(logging_config)
    if asynchronous:
        setup_logging_queue()


def init(path=None, key="log_config", Loader=ConfigLoader, **kwargs):
    """Read configuration from a file, initialize logging and embed the logging
    configuration under ``key``.

    :param path: Path of the file from which to read the configuration.  If
        None, the path is read from the value of the ``CONFIG`` environment
        variable.  If no such variable, path defaults to ``config.yaml``.
    :param key: The key under which to find the path to the logging
        configuration.
    :param Loader: See :func:`miniscule.load_config`.
    :param kwargs: Passed directly to :func:`miniscule.logs.init_logging`.

    :returns: See :func:`miniscule.load_config`.
    """
    config = read_config(path, Loader)
    if "logging" not in config and config.get(key):
        config["logging"] = _read_logging_config(config.pop(key))
    if "logging" in config:
        _set_up_logging(config["logging"], **kwargs)
    return config



import re

import setuptools


def read_file(path):
    with open(path, "r") as handle:
        return handle.read()


def read_version():
    try:
        s = read_file("VERSION")
        m = re.match(r"v(\d+\.\d+\.\d+(-.*)?)", s)
        return m.group(1)
    except FileNotFoundError:
        return "0.0.0"


long_description = read_file("docs/source/description.rst")
version = read_version()

setuptools.setup(
    name="miniscule",
    description="""
    A YAML based, pluggable configuration library inspired by Aeson""",
    keywords="configuration",
    long_description=long_description,
    include_package_data=True,
    version=version,
    url="https://gitlab.com/bartfrenk/miniscule/",
    author="Bart Frenk",
    author_email="bart.frenk@gmail.com",
    package_dir={"miniscule": "src/miniscule"},
    entry_points={"console_scripts": ["miniscule = miniscule:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    # TODO: Make aws resolvers dependent on existence of boto3 dependency
    install_requires=["PyYAML>=3"],
    data_files=[(".", ["VERSION"])],
    setup_requires=["pytest-runner"],
    tests_require=[
        "pytest>=4",
        "mock>=2.0.0",
        "boto3>=1.12.0,<2.0",
        "smart-open>=5.2.1,<6.0",
    ],
    packages=setuptools.find_packages("src"),
)



import json
import logging

import yaml

from miniscule.base import ConfigLoader

log = logging.getLogger(__name__)


try:
    import boto3
    import botocore

    IGNORED_EXCEPTIONS = (
        botocore.exceptions.ClientError,
        botocore.exceptions.NoCredentialsError,
    )
except ImportError:
    pass
else:

    def init():
        yaml.add_constructor(
            "!aws/sm", secrets_manager_constructor, Loader=ConfigLoader
        )
        yaml.add_constructor(
            "!aws/secret", secrets_manager_constructor, Loader=ConfigLoader
        )

    def _secrets_manager_client():
        return boto3.client("secretsmanager")

    def secrets_manager_constructor(loader, node):
        friendly_name = loader.construct_yaml_str(node)
        client = _secrets_manager_client()

        secret_string = None
        try:
            response = client.get_secret_value(SecretId=friendly_name)
            secret_string = response.get("SecretString")
        except IGNORED_EXCEPTIONS as exc:
            log.warning(
                "Unable to obtain secrets manager value for %s: %s", friendly_name, exc
            )
            return None
        return _maybe_parse_secret_string(secret_string)

    def _maybe_parse_secret_string(secret_string):
        if secret_string is None:
            return None
        return _maybe_parse_as_json(secret_string) or secret_string

    def _maybe_parse_as_json(secret_string):
        try:
            return json.loads(secret_string)
        except json.JSONDecodeError:
            return None

    init()
